/*

The MIT License (MIT)

Copyright (c) 2015 VeryPositive SIA

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

*/

'use strict';

(function (root) {

	// # Internals
	// If you're looking for the API, it's at the bottom.
	
	// ## Single module enforcement
	
	// Coastline relies on, among other things
	//
	// * recognizing Context objects as its own and treating them specially
	// * single processing loop
	//
	// Therefore there must only be one Coastline per VM for things to go smoothly.
	// Since v0 and v1 are totally incompatible, we now use `CL` instead of `coastline` to distinguish them.
	
	var version = [5, 1, 2];
	var globalObject = typeof window != 'undefined' ? window : typeof global != 'undefined' ? global : {};
	if (globalObject.CL) {
		var instead = globalObject.CL;
		if (instead.version[0] != version[0] || instead.version[1] != version[1] || instead.version[2] < version[2]) {
			typeof console != 'undefined' && console.log('Coastline: using version ' + instead.version.join('.') + ' instead of ' + version.join('.') + '. Expect bugs. See https://www.npmjs.org/package/coastline#use-in-modules');
		}
		if (typeof module != 'undefined') {
			module.exports = instead;
		}
		else {
			root.CL = instead;
		}
		return;
	}
	
	// ## Utility functions
	
	// Having global classes/functions in a variable increases performance by an order of magnitude in V8.
	// Yes, even if you only use them in if's that should never execute in the first place.
	// I would guess it has something to do with branch prediction and CPU cache and other scary stuff.
	var ErrorClass = Error;
	var ArrayClass = Array;
	var setTimeoutFun = setTimeout;
	var setImmediateFun = typeof setImmediate == 'function' ? setImmediate : function (fn) { setTimeoutFun(fn, 0) };
	
	// > *...it is deliberate that we don't expose a way to detect whether a function is a generator or not.
	// > The reason is that any function can return an iterable object so it does not matter if it is a function or a generator function.*
	// 
	// Unfortunately it matters to me (since I need to know whether the user meant to return a value or run a generator).
	// I guess I'll have to stick to checking that obj is both an iterator and iterable.
	var isGenerator;
	if (typeof Symbol != 'undefined' && Symbol.iterator) {
		var iteratorSymbol = Symbol.iterator;
		isGenerator = function (obj) {
			return obj && typeof obj[iteratorSymbol] == 'function' && typeof obj.next == 'function';
		}
	}
	else {
		isGenerator = function () { return false; };
	}
	
	var isArray;
	if (ArrayClass.isArray) {
		isArray = ArrayClass.isArray.bind(ArrayClass);
	}
	else {
		isArray = function (arg) {
			return Object.prototype.toString.call(arg) === '[object Array]';
		};
	}
	
	var hasOwnPropertyFunction = Object.prototype.hasOwnProperty;
	var hasOwnProperty = function (obj, key) {
		return hasOwnPropertyFunction.call(obj, key);
	};
	
	var isFunSymbol = typeof Symbol != 'undefined' ? Symbol('isCLFun') : '@isCLFun';
	
	var now = Date.now ? Date.now : function () {
		return new Date().getTime();
	};
	
	// ## Main task processing loop.
	
	// This is used to avoid stack overflows.
	// It also allows to do something after `.enqueue()` call but before the next tasks starts.
	// This doesn't seem to have any performance impact (if not positive impact) so we use it at all times.
	//
	// Also, against all logic push/shift on a single array seems to be the fastest way to do this in V8.
	// No bloody idea why. Maybe it's actually a linked list?
	var pushToQueue;
	(function () {
	
		var queues = {};
		for (var p = -5; p <= 5; p++) queues[p] = [];
		
		var process = function () {
		
			var p = maxPriority;
			do {
				if (queues[p].length) {
					queues[p].shift().run();
				}
				else {
					maxPriority--;
				}
				p = maxPriority;
			} while (p >= -5);
			
			// Past this point process() needs to be run again
			processing = false;
			
			// Resolve and clean up stuff
			resolveDeadlocks();
			if (debug && active.length > 100 && active.length > lastActive * 2) cleanActive();
			
		};
		
		var maxPriority = -5;
		var processing = null;
		
		pushToQueue = function (c) {
			queues[c.priority].push(c);
			if (c.priority > maxPriority) maxPriority = c.priority;
			if (!processing) {
				processing = true;
				setImmediateFun(process);
			}
		};
		
	})();
	
	
	// ## Root object
	// Also a function. I was not sure what the default behaviour of CL() should be.
	// Earlier coastline() created a context object, but we don't need user to work with contexts directly anymore.
	// So instead it only does that when no args are given. Just in case.
	var CL = function (arg) {
		if (typeof arg == 'function') return CL.fun(arg);
		else if (arg) return CL.obj(arg);
		else return new Context({});
	};
	
	// Coastline version, as array
	CL.version = version;
	
	// Key used to bind Contexts to objects. This is dirty but works everywhere on every object you'd want to bind to.
	// I might use a WeakMap for this later but that would mean different side effects depending on whether the browser
	// supports this feature or not. I'm not sure I want that.
	CL.key = typeof Symbol != 'undefined' ? Symbol('CL') : '@CL';
	
	// ## Global private variables
		
	// Current context, that is the context whose method or generator is currently being run.
	var current;
	
	// A context that is being built, by `cl.cb()` and friends.
	// When current yields, this will be considered the yielded value instead of whatever the argument to yield really
	// was and pushed to current context.
	// If `CL.bg()` is run when a building context exists, it will be run as a background task.
	var building;
	
	// The id is only used for debugging purposes but I leave it in just in case.
	var nextTaskId = 1;
	
	// A unique object that will be returned by `iterate()` and `apply()` if an error is caught.
	var errorMarker = {};
	
	// Map of tasks that are waiting on hosts, used for deadlock resolution
	var waiting = {};
	
	// ## Internal debugging stuff
	
	// Whether the debugging is on
	var debug = false;
	
	// Whether logging is on
	var log = false;
	
	// Logged contexts array
	var logEntries;
	
	// Active Contexts
	var active = [];
	
	// How many actually active contexts there were the last time
	var lastActive = 0;
	
	// Clean up active context list from finished contexts
	var cleanActive = function () {
		var clean = [];
		var had = {};
		for (var i = 0; i < active.length; i++) {
			var c = active[i];
			if (had[c.id] || c.finished) continue;
			clean.push(c);
			had[c.id] = true;
		}
		active = clean;
		lastActive = active.length;
	};
	
	// ## Context object
	// The core of Coastline. Context objects represent a context of execution, a stack frame if you wish.
	// Most of the functionality is implemented here.
	// Most of the methods of Context objects are private to the library and should not be called directly.
	var Context = CL.Context = function (options) {
		
		// This can be used for debugging.
		this.id = nextTaskId++;
		
		// Error object to gather stack data from
		this.errorObject = debug ? new ErrorClass() : null;
		
		// If logging is enabled, add us to list and set created time
		if (log) {
			this.createdTime = now();
			logEntries.push(this);
		}
		
		// Queue is where all the subtasks are put. Queue starts execution once the main task is done.
		// The tasks are removed from queue once they start. Caller-replacing tasks bypass the queue (see `resolve` below)
		// In case of generators, the whole queue is executed every time a generator yields.
		// 
		// Initializing this to [] has a significant performance impact in V8, so we don't.
		this.queue = null;
		
		// Caller is the Context that is responsible for the creation and execution of this Context.
		// Caller also waits for this Context to finish and will retrieve and use it's `.value` when it does.
		// Tasks with a caller may not be restarted (see `enqueue` for explanation)
		// This will be set by `.push()`
		this.caller = null;
		
		// Preset host Context to null. This used to be called "parent" in v0 but I think "host" suits better.
		// Host context is the main race condition control mechanism of Coastline.
		// Host context is created and bound to an object when `CL.obj` is run or a `CL.fun` is run as an object's method.
		// When a task has a host, it will first reach it's time in caller's queue, then add itself to the host's queue.
		// Host objects should not have a caller so they can be restarted any number of times.
		// Since 5.0 this is initially null and is only set in resolve()
		this.host = null;
		
		// Object the host of which we should use
		this.hostObject = options.hostObject || null;
		
		// An object the host context is bound to.
		this.bind = options.bind || null;
		
		// What gets put into `this` for the context's method/generator. Generally used by `CL.fun()` methods.
		this.object = options.object || null;
		
		// The function that is run when the context reaches it's place in a queue (or is run for any other reason).
		// If this is a generator function, it's result is assigned to `.generator` and the context is re-run.
		// See `runMethod`
		this.method = options.method || null;
		
		// The generator/iterator that is either assigned directly from yield value, or by running a generator .method.
		// See `runGenerator`
		this.generator = options.generator || null;
		
		// A Promises/A+ object assigned from a yield value. *development in progress*
		// See `runPromise`
		this.promise = options.promise || null;
		
		// Arguments to the method. Assigned by `CL.fun()`
		this.args = options.args || null;
		
		// If this is false, the Context will be Done as soon as method/generator finishes.
		// Otherwise, it will wait for callback return previously by `CL.done()` to be called.
		// Used for Contexts built by `CL.done()`/`CL.cb()`/`CL.error()`
		// See `CL.done`, `runGenerator`
		this.working = options.working || false;
		
		// Is this a background task?
		// Background tasks do not block the caller
		// The caller still waits for them to finish before reporting itself	as done.
	 	this.bg = options.bg || false;
	 	
	 	// List of child `bg` tasks
	 	this.bgTasks = null;
		
		// Predefine all fields purely for optimization purposes. Many guides mention this and it seems to work.
		this.data = options.data || null;
		
		// Environment ancestor, treated same as caller for `CL.get()` purposes
		this.env = options.env || null;
		 
		// References to other contexts
		this.yielded = this.current = this.replaced = null;
		
		// Various execution states.
		this.enqueued = this.started = this.finished = this.isDone = this.paused = this.retrying = false;
		this.hostPushed = this.hostReached = false;
		this.methodDone = this.generatorDone = false;
		
		// Context return value. This is value of `return` or last `yield` by generator/method.
		// `null` is a valid value, so we use `undefined` to show lack thereof.
		this.value = void 0;
		
		// Error object. When an error is thrown, this is set in all the call tree until a `.catch` is found.
		this.error = null;
		
		// An error that occured to this context before it even started
		this.discard = null;
		
		// Error handlers. See `try` for more info.
		this.catch = options.catch || null;
		
		// Function to call when retrying
		this.retry = options.retry || null;
		
		// Event handlers
		this.handlers = null;
		
		// Priority, from -5 to 5
		this.priority = 0;
		
	};
	Context.prototype = {
	
		// ### .push()
		// Push a subtask Context to the queue. Used for both callers and hosts.
		push: function (c) {
			// If we are not the host, then we must be the caller - otherwise what is this Context doing here?
			if (!c.caller && c.host != this) {
				c.caller = this;
				c.priority = this.priority;
			}
			
			// Push to queue.
			if (!this.queue) this.queue = [c];
			else this.queue.push(c);
			
			if (c.host == this && this.bind && this.bind[CL.key] != this) internalError('Pushed onto removed host');
			
			// If we're a host and have a current, add this task to the list of waiting
			if (this.current && c.host == this) waiting[c.id] = c;
			
			// If we don't have a caller (therefore we are either a host or a non-host root Context created by CL.run())
			// and we haven't started yet (or have finished) then start us (and therefore start the pushed task afterwards)
			else if (!this.caller && !this.started) this.enqueue();
			
			// Return the pushed context
			return c;
		},
		
		// ### .runGenerator()
		// Run *an iteration* of the generator 
		runGenerator: function () {
			// Set `current` to current Context so that `CL.push()` and others know where to push.
			// In v0 you had to pass the context around manually and it was a lot of pain.
			// This is simpler and prevents unpleasant errors from passing the wrong `c` in the wrong place.
			current = this;

			// Run an iteration of the generator.
			// If we had previously yielded a context (or something that was converted to a context)
			// then pass its value, otherwise pass whatever was yielded.
			// (so that you can yield a method call even if it might not be a `CL.fun`)
			var ret = iterate(this.generator, this.yielded instanceof Context ? this.yielded.value : this.yielded);
			
			// Now the code has finished and there is no current context at this moment. `CL.push()` will create
			// a new Context if run now.
			current = null;
			
			// If an error was caught, stop processing this context forever (once a context errs, it can't be restarted).
			// This `return` effectively makes coastline forget about *this* context forever.
			// If we start keeping a list of active contexts for deadlock checks later, we'll need to clean it up here.
			if (ret == errorMarker) return;
			
			// Generator has finished
			if (ret.done) {
				this.generatorDone = true;
			}
			
			var isReturn = ret.done && ret.value !== void 0;
			var retval = ret.value;
			var rettype = typeof retval;
			
			// Whether to pause execution of the context due to a *bg* task.
			var pause = false;
			
			// If building a context via `CL.cb()` and friends, take this context as the yield value and restart building.
			if (building) {
				this.yielded = this.push(building);
				building = null;
			}
			
			// If it evaluates to `false` or is not an object or function,
			// then it's certainly nothing of our interest, so we skip extra checks.
			else if (!retval || (rettype != 'object' && rettype != 'function')) {
				this.yielded = retval;
			}
			
			// If it's a context, set `yielded` to it. On the next generator run its value will be used.
			else if (retval instanceof Context) {
				
				this.yielded = retval;
				
				// If it's a `bg` task or an unrelated task *and* it's not finished yet, wait for it to finish
				if ((retval.bg || retval.caller != this) && !retval.finished) {
					
					// If there was an error, handle it and stop processing *this* context
					if (retval.error) {
						
						this.handle(retval.error);
						
						// Forget about the context, see above.
						return;
					}
					
					// Pause the context, see below.
					pause = true;
					
					// Add self to the list of contexts that will be resumed when the *bg* task finishes
					(retval.handlers || (retval.handlers = [])).push(this);
				}
				
			}
			
			// If it was a `return` and not `yield`, consider functions, generators and then-containing-objects a value
			else if (isReturn) {
				this.yielded = retval;
			}
			
			// If yielded a generator (by a call to an unwrapped generator function) create and push a new context.
			else if (isGenerator(retval)) {
				this.yielded = this.push(new Context({ generator: retval }));
			}
			
			// If yielded a function, create and push a new context for this function.
			else if (rettype == 'function') {
				this.yielded = this.push(new Context({ method: retval }));
			}
			
			// If yielded a Promises/A+ object, process it via `wrapPromise`
			else if (typeof retval.then == 'function') {
				this.yielded = this.push(wrapPromise(retval));
			}
			
			// If all else fails, just pass through whatever it is.
			else {
				this.yielded = retval;
			}
			
			// Enqueue for next iteration or to mark ourselves as done -
			// unless waiting on a *bg* task. In that case we will be resumed via `handlers` array of that task.
			if (!pause) this.enqueue();
		},
		
		// ### .runMethod()
		// Runs the method passed to the Context. This includes generator functions that will return a generator
		// which will then be used by runGenerator.
		runMethod: function () {
			
			// Set `current` to self (see `runGenerator`)
			current = this;
			
			// Run method via a separate function (to reduce unoptimized code due to `try`)
			var retval = apply(this.method, this.object, this.args);
			
			// Set `current` back to null
			current = null;
			
			// If an error happened, forget about this context (see `runGenerator`)
			if (retval == errorMarker) return;
			
			// If the method returned another context, assign it to `.yielded` so we obtain the value on finish
			if (retval instanceof Context) this.yielded = retval;
			// If the method returned a generator (*not* a `function*`, but a result of it), set it as our `.generator`
			else if (isGenerator(retval)) this.generator = retval;
			// Otherwise just set the return value of the context.
			else if (this.value === void 0) this.value = retval;
			
			// Mark us as having run the method so we don't return here.
			this.methodDone = true;
			
			// Enqueue for any subsequent actions needed
			this.enqueue();
		},
		
		// ### .runQueue()
		// Gets a single Context out of queue and runs it.
		runQueue: function () {
			var c;
			if (this.queue.length < 1000) {
				c = this.queue.shift();
			}
			else {
				if (!this.queueHead) this.queueHead = 0;
				c = this.queue[this.queueHead++];
				if (this.queueHead >= this.queue.length) {
					this.queue = [];
					this.queueHead = 0;
				}
				else if (this.queueHead >= 1000) {
					this.queue = this.queue.slice(this.queueHead);
					this.queueHead = 0;
				}
			}
			
			// If we're the host
			if (c.host == this) {
				// then mark the task as having reached the host
				c.hostReached = true;
				// remove it from waiting task list if it was there
				delete waiting[c.id];
				// set it as our current task (no relation to global `current`)
				// (note that when executes it will be current both to host and to caller)
				this.current = c;
				// ...and enqueue it
				c.enqueue();
			}
			
			// If the task has not been marked as *bg*, start it.
			else if (!c.bg) {
				this.current = c;
				c.enqueue();
			}
			
			// Otherwise it's a *bg* task. They end up in queue before being marked as *bg*
			// because `CL.bg()` may be run on an already pushed context.
			// Just ignore it and keep processing the queue.
			else {
				this.enqueue();
			}
		},
		
		// ### .unpackMethod()
		// Unpacks method passed as array [object, method] where method can either be a Function or an object's method name.
		// Also sets `.host` and `.retry` if the method is a wrapped function.
		unpackMethod: function () {
			var arr = this.method;
			this.object = arr[0];
			this.method = typeof arr[1] == 'string' ? this.object[arr[1]] : arr[1];
			if (this.method[isFunSymbol]) {
				if (!this.method.loose) this.hostObject = this.object;
				this.retry = this.method.retry;
			}
			this.enqueue();
		},
		
		// ### .run()
		// Context's main entry point. This is called from the main loop and decides what to do depending on state.
		run: function (force) {
		
			// This can happen if a task was executed out of band (`CL.immediate()`)
			if (!this.enqueued && !force) return;
			
			// Should never happen in theory, but never rely on theory alone.
			if (this.current) internalError('run while having current');
			
			// Allow ourselves to be enqueued again.
			this.enqueued = false;
			
			// This can happen if a task was paused after being enqueued
			if (this.paused) return;
			
			// This can happen when a *bg* task errors out after the caller is enqueued (e.g. in `CL.parallel()`)
			if (this.error) return;
			
			// Announce the error in case we failed way too early
			if (this.discard) return this.throw.apply(this, this.discard);
			
			// If we have a host, call `.resolve()` to push and resolve any recursion.
			if (this.hostObject && !this.hostPushed) this.resolve();
			
			// If we've already pushed to a host and are waiting, do nothing.
			else if (this.hostPushed && !this.hostReached); // noop
			
			// If we were passed a member method, unpack it
			else if (this.method && !this.methodDone && isArray(this.method)) this.unpackMethod();
			
			// If we have an unprocessed method, queue or generator, run them.
			else if (this.method && !this.methodDone) this.runMethod();
			else if (this.queue && this.queue.length) this.runQueue();
			else if (this.generator && !this.generatorDone) this.runGenerator();
			
			// If we have background tasks, wait for them
			else if (this.bgTasks && this.bgTasks.length) this.wait();
			
			// If we're retrying, free the host and add ourselves to its queue
			else if (this.retrying) this.tryAgain();
			
			// If we're already marked as done but not finished, call `.finish()`
			else if (this.isDone && !this.finished) this.finish();
			
			// Otherwise if not marked as `working` (by being a built context or via manual `CL.work()`)
			// mark ourselves as done.
			else if (!this.isDone && !this.working) this.done();
		},
		
		// ### .wait()
		// Wait for background tasks to finish.
		wait: function () {
			// Loop over background tasks
			var tasks = this.bgTasks;
			for (var i = 0; i < tasks.length; i++) {
				var c = tasks[i];
				// If there was an error in a bg task, fail too
				if (c.error) {
					this.handle(c.error);
					return;
				}
				// If it's not finished, add ourselves to the list of handlers listeners and stop doing anything
				if (!c.finished) {
					(c.handlers || (c.handlers = [])).push(this);
					if (i) this.bgTasks = tasks.slice(i);
					return;
				}
			}
			this.bgTasks = [];
			this.run(true);
		},
		
		// ### .clearBg()
		// Clear finished background tasks from the list. Used by `CL.parallel()` etc. Does not modify execution state.
		clearBg: function () {
			// Loop over background tasks
			var tasks = this.bgTasks;
			var left = [];
			for (var i = 0; i < tasks.length; i++) {
				var c = tasks[i];
				// If it's not finished, add ourselves to the list of handlers listeners and stop doing anything
				if (!c.finished) {
					left.push(c);
				}
			}
			this.bgTasks = left;
		},
		
		// ### .tryAgain()
		// Free the host, add ourselves back to queue and run again
		tryAgain: function () {
			// Don't retry again if everything goes well
			this.retrying = false;
			
			// If we have a host, free it
			if (this.host) {
				this.host.currentFinished(this);
			}
			
			// Reset state
			this.methodDone = this.generatorDone = this.hostPushed = this.hostReached = false;
			this.enqueue();
		},
		
		// ### .done()
		// Marks the task as done and optionally sets the return value.
		// Can also be called manually on a "working" task.
		done: function (value) {
			// Calling `.done()` twice is an error, and we naively believe it will be made by the user and not ourselves.
			if (this.isDone) userError('task already done');
			// If we don't have a value yet, use whatever was passed (even if nothing was)
			if (typeof value !== 'undefined') this.value = value;
			// Set the flag
			this.isDone = true;
			// Call `.run()` directly. Since `.done()` is only ever called once, this shouldn't lead to stack problems.
			this.run(true);
		},
		
		// ### .finish()
		// Marks the task as finished and informs every context interested in the fact.
		finish: function () {
			// `.finish()` should not be called twice either, but if it is, it's certainly our fault.
			if (this.finished) internalError('task already finished');
			
			// Set flags
			this.finished = true;
			this.started = false;
			
			if (log && !this.finishedTime) this.finishedTime = now();
			
			// If we still don't have a value, use the value of a yielded context.
			// Note that we need both this code and the code in `runGenerator` for it to work in every case.
			if (typeof this.value == 'undefined') {
				this.value = this.yielded instanceof Context ? this.yielded.value : this.yielded;
			}
			if (this.value && this.value instanceof Value) {
				this.value = this.value.value;
			}
			
			// If we have a host, and we were running, then we're its current.
			if (this.host) {
				this.host.currentFinished(this);
			}
			
			// If we have a caller, inform them
			if (this.caller) {
				// If it's not a *bg* task, then it must be current
				if (!this.bg) this.caller.currentFinished(this);
			}
			
			// Inform any other interested parties that we've finished
			if (this.handlers) for (var i = 0, l = this.handlers.length; i < l; i++) {
				var hdl = this.handlers[i];
				hdl.handleFinish(this);
			}
			this.handlers = null;
			
			// If we are a host, remove ourselves from the object to avoid retaining everything forever
			if (this.bind) this.bind[CL.key] = null;
			
		},
		
		// ### .currentFinished()
		// Called by a child task to inform the caller that it's finished.
		currentFinished: function (c) {
		
			// Calling task must be our current
			if (this.current != c) internalError('current not c when finished');
			
			// If we're a host and something was replaced by `.resolve()`, replace it back. It will continue
			// because it will also receive a `.currentFinished()` since it must be in a caller chain for `c`.
			if (this.replaced && this.replaced.length) {
				this.current = this.replaced.pop();
			}
			
			// Otherwise set our current to null and keep processing queue or generator or whatever.
			else {
				this.current = null;
				this.enqueue();
			}
		},
		
		// ### .enqueue()
		// Places the task in the main loop queue.
		enqueue: function () {
			
			// If we're already in the queue or are manually paused, do nothing.
			if (this.enqueued || this.paused) return;
			
			// This can happen e.g. if you try to manually push to some context.
			if (this.finished && this.caller) userError('restarting a context with a caller');
			
			// This shouldn't ever happen since we forget about errored contexts immediately.
			if (this.error) internalError('restarting an errored context');
			
			// Debug: If we haven't started yet or had finished, push to active contexts list
			if (debug && (!this.started || this.finished)) active.push(this);
			
			// Set flags
			this.enqueued = true; 
			this.started = true;
			this.finished = false;
			
			if (log && !this.startedTime) this.startedTime = now();
			
			// Call the global method to push to global queue.
			pushToQueue(this);
		},
		
		// ### .resolve()
		// Place the task in the host's queue and resolve a recursive call via the same host.
		resolve: function () {
		
			this.host = CL.obj(this.hostObject);
		
			// Yep, this is not supposed to happen either.
			if (this.host == this.caller) internalError('host is also caller');
			
			// Set flag
			this.hostPushed = true;
			
			// If the host has a current, it might be one of our callers, so we replace it to avoid deadlocking.
			var pcur = this.host.current;
			if (pcur) {
				// Iterate over our caller chain
				var cur = this;
				do {
					if (cur == pcur) {
						// If one of our callers is host's current, replace it and return.
						this.host.replace(this);
						return;
					}
				} while (cur = cur.caller);
			}
			
			// If not returned above, push self to host's queue.
			this.host.push(this);
		},
		
		// ### .replace()
		// Called by `.resolve()` on a host Context to replace old `.current` with a new one.
		replace: function (c) {
			// This array will keep the full stack of replaced contexts (there might be more than one)
			this.replaced = this.replaced || [];
			this.replaced.push(this.current);
			this.current = c;
			// This is an alternative way to reach a host without ever going into queue
			c.hostReached = true;
			c.enqueue();
		},
		
		// ### .then(success, error)
		// Implements Promises/A+ support. Note that calling this on a top context will also prevent it from crashing Coastline.
		then: function (succ, err) {
			var c = this;
			var cc;
			// If there already was an error, run error handler immediately
			if (c.error) {
				cc = new Context({
					method: function* () {
						if (err) yield err(c.error);
						else throw c.error;
					}
				});
				cc.enqueue();
			}
			// If we're already finished, run success handler immediately
			else if (c.finished) {
				cc = new Context({
					method: function* () {
						if (succ) yield succ(c.value);
						else return c.value;
					}
				});
				cc.enqueue();
			}
			// Else create a context with overloaded error/finish handlers
			else {
				cc = new Context({});
				cc.handleError = function (c, e) {
					if (err) {
						try {
							this.value = err(e);
							this.finish();
						}
						catch (e) {
							this.handle(e);
						}
						return true;
					}
					else {
						this.finish();
						return false;
					}
				};
				cc.handleFinish = function (c) {
					if (succ) {
						try {
							this.value = succ(c.value);
							this.finish();
						}
						catch (e) {
							this.handle(e);
						}
					}
					else {
						this.value = c.value();
						this.finish();
					}
				};
				(c.handlers || (c.handlers = [])).push(cc);
			}
			return cc;
		},
		
		// ### .pause()
		// Manually pause the context. Used by `CL.parallel()` and friends.
		pause: function () {
			this.paused = true;
		},
		
		// ### .resume()
		// Manually resume the context after `.pause()`. Safe to be called at any moment.
		resume: function () {
			this.paused = false;
			if (!this.current && !this.enqueued) this.enqueue();
		},
		
		// ### .throw()
		// Prepare and throw an error from this context.
		throw: function (type, error) {
			// Prepare an Error object from given parameters, then handle it recursively.
			this.handle(prepareError(type, error));
		},
		
		// ### .handle()
		// Recursively handles the error.
		handle: function (e) {
			var i, l;
			var handled = false;
			
			// If we're the first `.handle()` call to encounter this Error object, set ourselves as being at fault.
			if (!e.CL) {
				e.CL = this;
			}
			
			// If we're a catch context, try handling the error.
			var hdl;
			if (this.catch && (
				// First try the specific type, then a wildcard. CL errors do not match the wildcard.
				(hdl = this.catch[e.type]) || ((hdl = this.catch['*']) && e.type != 'CLError' && e.type != 'CLInternalError')
			)) {
				// If the *error handler* errors out too, don't call it again.
				this.catch = null;
				// Whatever we had as `.current` is gone forever now, but we have work to do.
				this.current = null;
				// Also drop the queue
				this.queue = [];
				// Push the error handler as a new child Context to self.
				this.push(this.yielded = new Context({
					method: hdl,
					args: [e]
				}));
				// Merrily carry on
				this.enqueue();
				
				return;
			}
			
			if (this.retry && e.type == 'CLDeadlock') {
				
				// Make sure we don't continue running anything right now
				// FIXME background tasks will keep running, not sure what to do about this.
				this.methodDone = this.generatorDone = true;
				this.queue = [];
				this.current = null;
				
				// Create an array of arguments for the handler
				var args = [e];
				for (i = 0; i < this.args.length; i++) {
					args.push(this.args[i]);
				}
				
				// Push the error handler as a new child Context to self, unless "true" was passed instead of a handler.
				if (this.retry !== true) this.push(new Context({
					method: this.retry,
					object: this.object,
					args: args
				}));
				
				// Check that there is no offending host in our callers
				// This would mean that we don't release anything and can keep retrying forever 
				var found = false;
				var caller = this.caller;
				do {
					if (caller.host == e.host) found = true;
				} while (caller = caller.caller);
				
				// If we're safe to retry, do that
				if (!found) {
					// Set a special flag that will free the host but not the caller
					this.retrying = true;
					// Carry on
					this.enqueue();
				}
				else {
					// Do not retry
					this.retry = null;
					// If there's a handler, allow it to run
					if (this.retry !== true) {
						// and handle the error again after it does
						this.push(new Context({
							method: function () {
								CL.throw(e);
							}
						}));
						// Continue
						this.enqueue();
					}
					// Otherwise just handle again
					else {
						this.handle(e);
					}
				}
				
				return;
			}
			
			// If we're not a catch or retry context, accept our fate
			// Set `.error` to make sure we're not ever run again.
			this.error = e;
			
			// If we have a host...
			if (this.host) {
				// ...and we were its current, let it continue with other tasks
				if (this.host.current == this) this.host.currentFinished(this);
				// ...otherwise remove us from the waiting list
				else delete waiting[this.id];
			}
			
			// Otherwise if we're not the last in the caller chain, kill ourselves with a bang.
			if (this.caller) {
				// If we're not a bg task, pass the error up the chain.
				// Also pass if caller is paused (as used in `CL.parallel()`) since it might as well never resume.
				// If caller has already errored before, fail silently. (maybe there should be CL.onBgError for this?)
				if ((!this.bg || this.caller.paused) && !this.caller.error) {
					this.caller.handle(e);
				}
				handled = true;
			}
			
			// Cycle through handlers looking for any Contexts waiting on us.
			// These must also handle the error, since it's either that or a deadlock.
			if (this.handlers) for (i = 0, l = this.handlers.length; i < l; i++) {
				var fhdl = this.handlers[i];
				if (!fhdl.error) handled = fhdl.handleError(this, e) || handled;
			}
			this.handlers = null;
			
			// If the error was not handled in any way, fail hard
			if (!handled) {
				
				// Behavior after this is currently undefined.
				// Basically, don't have unhandled errors in production. 
				// Thank you, and have a nice day.
				
				CL.onError(e);
				throw e;
				
			}
		},
		
		// ### .handleFinish()
		// Called when a Context we are waiting for finishes. Overridden in `.then()` and `Queue`.
		handleFinish: function (c) {
			// Do nothing if we had an error (can still find ourselves here in CL.parallel)
			if (this.error) return;
			// Resume our execution
			this.resume();
		},
		
		// ### .handleError()
		// Called when a Context we are waiting for has an error. Overridden in `.then()` and `Queue`.
		handleError: function (c, e) {
			this.handle(e);
			return true;
		},
		
		trace: function () {
			if (!debug) return 'No Coastline stack available, enable with CL.debug(true);';
			var c = this;
			var stack = [];
			do {
				// Provide some basic context info
				var header = 'Context #' + c.id + ' ';
				if (c.tag) header += '(' + c.tag + ') ';
				if (c.method) header += c.method.displayName || c.method.name;
				stack.push(header);
				// Debug was disabled when this context was created
				if (!c.errorObject) {
					stack.push('* no errorObject');
					continue;
				}
				// This will happen in IE
				if (!c.errorObject.stack) {
					stack.push('* no stack trace available');
					continue;
				}
				var split = c.errorObject.stack.split('\n');
				// Skip lines that are not stack trace (e.g. error description)
				var i = 0;
				for (; i < split.length; i++) {
					if (split[i].indexOf('Context') >= 0) break;
				}
				// Is there anything remaining?
				if (!split[i]) {
					stack.push('* no stack trace available');
					continue;
				}
				// Skip the coastline.js line
				i++;
				// Always push the first line
				stack.push('  ' + split[i++].trim());
				// Find all lines that were not from coastline
				for (var n = 0; i < split.length; i++, n++) {
					var str = split[i].trim();
					if (str.match(/Context(\.prototype)?\.run/) || str == 'at [object Generator].next (native)') break;
					stack.push('  ' + str);
				}
			} while (c = c.caller);
			return stack.join('\n');
		},
		
		set: function (key, val) {
			var data = (this.data || (this.data = {}));
			data[key] = val;
		},
		
		get: function (key) {
			var c = this;
			while (c) {
				if (c.data && key in c.data) {
					return c.data[key];
				}
				c = c.caller || c.env;
			}
		}
		
	};
	
	// ## Global private methods
	
	// ### iterate()
	// Calls an iterator's `.next()` method and catches any immediate errors.
	var iterate = function (gen, arg) {
		try {
			return gen.next(arg);
		}
		catch (e) {
			if (current) current.throw(e);
			else throw e;
			return errorMarker;
		}
	};
	
	// ### apply()
	// Calls a method and catches any immediate errors.
	var apply = function (fun, ctx, args) {
		try {
			return fun.apply(ctx, args);
		}
		catch (e) {
			if (current) current.throw(e);
			else throw e;
			return errorMarker;
		}
	};
	
	// ### getBuilding()
	// Creates or returns a building context, that is the context that will be pushed to current on the next `yield`.
	var getBuilding = function () {
		if (!current) userError('no building context present');
		if (!building) building = new Context({ working: true });
		return building;
	};
	
	// ### builtDone()
	// Run by `CL.cb()` and `CL.done()` created callbacks.
	var builtDone = function (b, val) {
		// If the callback was not run immediately and the task started, mark it as done.
		if (b.started) {
			b.done(val);
		}
		// Otherwise mark it as not working so it starts and finishes immediately,
		else {
			b.working = false;
			b.value = val;
		}
	};
	
	// ### wrapPromise()
	// Creates a context that will either be done when the promise resolves, or will throw an error if it rejects.
	var wrapPromise = function (promise) {
		var c = new Context({ working: true });
		promise.then(function (val) {
			builtDone(c, val);
		}, function (err) {
			c.throw(err);
		});
		return c;
	};
	
	// ### prepareError()
	// Prepares an Error object from given parameters.
	var prepareError = function (type, error) {
		var e;
		// If we have two parameters:
		if (error) {
			// ...and the second one is an actual Error, just use that.
			if (error instanceof ErrorClass) {
				e = error;
			}
			// ...otherwise make a new one and use whatever was passed for message or data
			else {
				if (typeof error == 'string') {
					e = new ErrorClass(error);
				}
				else {
					e = new ErrorClass();
					e.data = error;
				}
			}
			// Set Error's (hopefully otherwise unused) `.type` property to explicitly provided `type`
			e.type = type;
		}
		// If we have one parameter:
		else {
			// ...and it's already an Error, use it
			if (type instanceof ErrorClass) {
				e = type;
				// if itt doesn't have a type defined, guess it
				if (!e.type) {
					// if it's a TypeError etc., use that
					if (e.name != 'Error') {
						e.type = e.name;
					}
					// otherwise hope it follows the standard patter of *ErrorType: error message*
					else {
						var ermm = e.message.match(/^(.*?):/);
						e.type = (ermm && ermm[1]) || 'Error';
					}
				}
			}
			// ...and it's not an Error, make a new one
			else {
				// if passed a string, use it as both type and message
				if (typeof type == 'string') {
					e = new ErrorClass(type);
					e.type = type;
				}
				// otherwise make a message-less Error with a default type
				else {
					e = new ErrorClass();
					e.data = type;
					e.type = 'Error';
				}
			}
		}
		// If there is a current, mark it as being at fault here
		if (!e.CL) {
			e.CL = current;	
		}
		return e;
	};
	
	// ### internalError()
	// Throws an internal error in case Coastline has no idea what's happening.
	var internalError = function (msg) {
		throw prepareError('CLInternalError', 'Coastline internal error: ' + msg);
	};
	
	// ### userError()
	// Throws a error at the user for doing something weird.
	var userError = function (msg) {
		throw prepareError('CLError', 'Coastline: ' + msg);
	};
	
	// ### resolveDeadlocks()
	// Resolves deadlocks occuring because of multiple chains having costs calling each other.
	// Resolves indirect cases too (A->B, B->C, C->A).
	var resolveDeadlocks = function () {
		// Map of all locked hosts
		var locks = {};
		// Go over the list of waiting tasks
		for (var cid in waiting) {
			if (!hasOwnProperty(waiting, cid)) continue;
			cid = parseInt(cid);
			var c = waiting[cid];
			// If c has no caller, it's certainly not blocking anything
			if (!c.caller) return;
			// c.host is what c is currently waiting on
			var hid = c.host.id;
			// Save the locks of the host to check against
			var hlocks = locks[hid] || {};
			// Save the Contexts we've been through in this iteration
			var traced = {};
			// Recursive function that checks all the caller chains that are locked by this task
			var trace = function (caller) {
				traced[caller.id] = true;
				do {
					// If caller has no host, it can't lock it
					if (!caller.host) continue;
					var ch = caller.host;
					var chid = ch.id;
					// If this host locks our host via a different caller chain, and we have locked it ourselves
					// then the only way out is to kill one of the Contexts.
					if (hlocks[chid]) {
						// Find the latest Context in the offending chain
						var oc = caller;
						while (oc.current) oc = oc.current;
						// Remove it from host's queue 
						oc.host.queue.splice(oc.host.queue.indexOf(oc), 1);
						var e = prepareError('CLDeadlock', 'Coastline: deadlock encountered');
						// Set the offending host
						e.host = caller.host;
						oc.throw(e);
						return true;
					}
					// Save information about this caller chain locking this host for future iterations of the loop
					if (!locks[chid]) locks[chid] = {};
					locks[chid][hid] = cid;
					// Iterate over all the Contexts waiting on this host after us in case we are locked via them
					for (var i = 0; i < ch.queue.length; i++) {
						var qc = ch.queue[i];
						if (!traced[qc.id]) {
							// If we found a deadlock elsewhere, return immediately
							if (trace(qc)) return true;
						}
					}
				} while (caller = caller.caller);
				return false;
			};
			// Run trace on tasks own caller
			// If found a deadlock, return so we don't crash other Contexts that might as well work it out now
			if (trace(c.caller)) return true;
		}
	};
	
	// # Coastline API
	
	// ### yield CL.push(Context c)
	// Explicitly push a context to currently running one or enqueue it as a top caller.
	// This may be made private in future if it doesn't find any use.
	var push = CL.push = function (c) {
		if (current) current.push(c);
		else c.enqueue();
		return c;
	};
	
	// ### CL.obj(Object)
	// Creates a host Context for the object and returns it.
	CL.obj = function (obj) {
		// Bind a Context to obj[CL.key] (`'@CL'` by default)
		if (!obj[CL.key]) obj[CL.key] = new Context({
			bind: obj,
			object: obj
		});
		return obj[CL.key]; 
	};
	
	// ### CL.fun(Function)
	// Wraps a function in another function that automatically sets its host to the object it's being called on.
	CL.fun = function (fun, options) {
		var wrapper = function () {
			// Collect arguments without screwing up V8 optimization
			var l, args = null;
			if (l = arguments.length) {
				args = new ArrayClass(l);
				for (var i = 0; i < l; i++) {
					args[i] = arguments[i];
				}
			}
			// Push a new Context
			return push(new Context({
				// This will be the `this` for the method
				object: this,
				// In non-strict browsers `this` will be `window`, but we don't want to bind to that
				hostObject: (!(options && options.loose) && this && this != root && this != CL) ? this : null,
				// Method
				method: fun,
				// Passed arguments, if any
				args: args,
				// Retry in case of deadlock
				retry: options && options.retry
			}));
		}
		wrapper[isFunSymbol] = true;
		wrapper.method = fun;
		wrapper.loose = options && options.loose;
		wrapper.retry = options && options.retry;
		return wrapper;
	};
	
	// ### CL.run(Method, args...)
	// Starts a new caller chain if the currently running code is not a part of a task already.
	// Can also be used to explicitly run something in current context (but usually `yield` would be enough).
	//
	// First argument can be a `CL.fun()`, a `function*` or a `function`.
	//
	// All the subsequent arguments are passed to the function.
	CL.run = function (fun) {
		var cd = {
			method: fun
		};
		// Collect arguments without screwing up V8 optimization
		var l = arguments.length - 1;
		if (l) {
			var args = new ArrayClass(l);
			for (var i = 0; i < l; i++) {
				args[i] = arguments[i + 1];
			}
			cd.args = args;
		}
		// Push a new Context
		return push(new Context(cd));
	};
	
	// ### CL.top(Method, args...)
	// Starts a new caller chain, regardless of when it's called. The arguments are the same as for `CL.run()`
	CL.top = function (fun) {
		var cd = {
			method: fun
		};
		// Collect arguments without screwing up V8 optimization
		var l = arguments.length - 1;
		if (l) {
			var args = new ArrayClass(l);
			for (var i = 0; i < l; i++) {
				args[i] = arguments[i + 1];
			}
			cd.args = args;
		}
		// Enqueue the context without pushing
		var c = new Context(cd);
		c.enqueue();
		return c;
	};
	
	// ### CL.spawn(Method, args...)
	// Starts a new caller chain but inherits the environment (`CL.set()`). The arguments are the same as for `CL.run()`
	CL.spawn = function (fun) {
		var cd = {
			method: fun,
			env: current
		};
		// Collect arguments without screwing up V8 optimization
		var l = arguments.length - 1;
		if (l) {
			var args = new ArrayClass(l);
			for (var i = 0; i < l; i++) {
				args[i] = arguments[i + 1];
			}
			cd.args = args;
		}
		// Enqueue the context without pushing
		var c = new Context(cd);
		c.enqueue();
		return c;
	};
	
	// ### yield CL.visit(Object, Method, args...)
	// Appends the task to the object's host context. The remaining arguments are the same as for `CL.run()`
	CL.visit = function (host, fun) {
		var cd = {
			method: fun,
			hostObject: host
		};
		// Collect arguments without screwing up V8 optimization
		var l = arguments.length - 2;
		if (l) {
			var args = new ArrayClass(l);
			for (var i = 0; i < l; i++) {
				args[i] = arguments[i + 2];
			}
			cd.args = args;
		}
		// Push a new Context
		return push(new Context(cd));
	};
	
	// ### yield CL.try(Function, Function | Object)
	// Your standard `try-catch`. `finally` is coming later.
	//
	// Handlers can either be a catch-all function (can be a generator or a `CL.fun()`)
	// or an object where key is error type and value is handler.
	//
	// `*` matches everything but internal errors.
	CL.try = function (fun, handlers) {
		// Create a separate context just for the error catching
		// (so that everything above terminates and everything below continues, potentially with other `.try()` calls)
		var cc = new Context({
			catch: (typeof handlers == 'function' || isArray(handlers)) ? { '*': handlers } : handlers
		});
		push(cc);
		// Push the Context with the method into it
		cc.push(cc.yielded = new Context({
			method: fun
		}));
		// Return the outer context. TODO think about return values
		return cc;
	};
	
	// ### CL.bg(any, args...)
	// Mark a task as a background task. The argument can be:
	//
	// * a building (external) task -- see `CL.cb()` below
	// * a newly created Context (e.g. via a call of `CL.fun()` result)
	// * a generator (result of a call of `function*`)
	// * a `function*` or a `function`
	// * a Promises/A+ object
	//
	// Background task will not block the caller but it will wait on it before finishing.
	// To wait manually and receive the return value, just save the result of `CL.bg()` to a variable
	// and `yield` it when you need it.
	CL.bg = function (fun) {
		var c;
		
		// If we have a building context, use that and disregard whatever is passed
		if (building) {
			c = building;
			c.bg = true;
			building = null;
		}
		// If passed a Context, mark it as `bg`
		else if (fun instanceof Context) {
			c = fun;
			c.bg = true;
		}
		// If passed a Generator, create a new Context for it
		else if (isGenerator(fun)) {
			c = push(new Context({
				generator: fun,
				bg: true
			}));
		}
		// If passed a promise, handle using `wrapPromise()`
		else if (typeof fun.then == 'function') {
			c = push(wrapPromise(fun));
			c.bg = true;
		}
		// Otherwise do what `CL.run()` does, except mark as `bg`
		else {
			var cd = {
				method: fun,
				bg: true
			};
			var l = arguments.length - 1;
			if (l) {
				var args = new ArrayClass(l);
				for (var i = 0; i < l; i++) {
					args[i] = arguments[i + 1];
				}
				cd.args = args;
			}
			c = push(new Context(cd));
		}
		
		// Push the Context to caller's bgTasks
		(current.bgTasks || (current.bgTasks = [])).push(c);
		
		// Enqueue the `bg` context in the main loop and return it
		c.enqueue();
		return c;
	};
	
	var throwFun = function (e) {
		throw e;
	};
	
	// ### CL.immediate(Context)
	// Try to execute and return the result of the context immediately (out-of-band).
	// Will also return a value of an already finished context.
	// If the Context has encountered an error previously, will throw it via native `throw`.
	// If immediate execution is not possible, will return the passed Context.
	CL.immediate = function (c, internal) {
		
		// Context has finished
		if (c.finished) return c.value;
		// Context has been paused
		if (c.paused) return c;
		// Context has encountered an error before
		if (c.error) throw c.error;
		
		// If there is a caller make sure `c` is not executed again
		var wasBg = c.bg;
		if (c.caller && c.caller.current != c) c.bg = true;
		
		// Either context is already enqueued (and therefore ok to run) or a set of acceptable conditions is met
		while (
			(c.enqueued || (c.method && !c.methodDone && (!c.hostObject || !c.host || c.hostReached || !c.hostPushed))) &&
			!c.current &&
			(!c.caller || !c.caller.current || c.caller.current == c)
		) {
			
			// Save current context
			var old = current;
			// Make sure the task executes at least once
			c.enqueued = true;
			// Run `.run()` until there's nothing left to do
			do c.run(); while (c.enqueued);
			// Restore current
			current = old;
			
			// If we have a subtask, try running it immediately too (avoid recursive immediate() for hosts)
			if (c.current && c.current.host != c) CL.immediate(c.current, true);
			
			// Context has encountered an error during execution
			if (c.error) throwFun(c.error);
			// Context has finished
			if (c.finished) return c.value;
			
			// If there is a host that has no current (and we might be first in queue), try calling immediate on that
			if (c.host && c.hostPushed && !c.host.current) {
				CL.immediate(c.host, true);
				if (c.host.current != c) break;
			}
		}
		
		if (!internal && !wasBg && current && current == c.caller) {
			c.bg = true;
			(current.bgTasks || (current.bgTasks = [])).push(c);
		}
		
		return c;
	}
	
	// ### CL.cur()
	// Returns the current Context. May be made private in the future.
	CL.cur = function () {
		return current;
	};
	
	// ### CL.priority(prio)
	// Set the priority of current Context, Valid values are -5 to 5 (inclusive).
	// Priorities are inherited from the caller. The default priority is 0.
	// Note that contexts with a higher priority are *always* executed first, regardless of how many are in the lower priority queues.
	CL.priority = function (prio) {
		current.priority = prio;
	};
	
	// ### CL.work()
	// Manually mark current Context as "working", that is, waiting for `.done()` to be called externally. May be made private in the future.
	CL.work = function () {
		current.working = true;
	};
	
	// ### CL.cb(errorType)
	// Creates a callback of form `callback(error, result)` that can be passed to any external function.
	// The task **must** yield immediately after said function is run so that this callback is bound to a correct object.
	//
	// Example: `yield fs.readFile('filename', CL.cb());`
	//
	// If an error is received, it will be thrown as via `CL.throw()` and the task will not continue.
	// Otherwise the result will be the return value of `yield`.
	// If `errorType` is given, Error's type will be set to it.
	CL.cb = function (errorType) {
		var b = getBuilding();
		return function (err, val) {
			// If error was not passed, mark the task as done
			if (!err) {
				builtDone(b, val);
			}
			// Otherwise throw. Give error type if passed, guess if not.
			else {
				var args = errorType ? [errorType, err] : [err];
				if (b.started) b.throw.apply(b, args);
				else b.discard = args;
			}
		}
	};
	
	// ### CL.done()
	// Same as `CL.cb()` but creates a `callback(result)` and does not handle errors.
	CL.done = function () {
		var b = getBuilding();
		return function (val) {
			builtDone(b, val);
		}
	};
	
	// ### CL.error(errorType)
	// Same as `CL.cb()` but creates a `callback(error)`.
	CL.error = function (type) {
		var b = getBuilding();
		return function (err) {
			var args = type ? [type, err] : [err];
			if (b.started) b.throw.apply(b, args);
			else b.discard = args;
		}
	};
	
	// ### CL.throw(type, error)
	// Throw an error up the stack. This error can be caught using `CL.try()`.
	// If it's not caught, Coastline will terminate and `CL.onError` will be run.
	// Can be run with 1 or 2 arguments:
	//
	// * string type, string error message (`type` will be assigned to Error's `.type`)
	// * string type, Error
	// * string type, any Object (will be assigned to `.data`)
	// * string type - will be both `.type` and `.message`
	// * Error
	// * any Object
	CL.throw = function (type, error) {
		throw prepareError(type, error);
	};
	
	// ### CL.onError
	// A callback that is run when an unhandled error reaches the top of the stack.
	// By default will print the Coastline stack and throw the error, therefore crashing the process in case of Node.js.
	// If you override it, be sure to crash the process too since the behavior of Coastline after an unhandled error
	// is currently undefined.
	//
	// The printed trace consists of tag (if present) full method code and arguments - it's the only useful information there is.
	// Previous version would create an Error object for every Context to get a proper stack, but that's way too slow.
	CL.onError = function (e) {
		console.log('');
		console.log(e.type);
		console.log('');
		console.log(e.message);
		console.log('');
		console.log(e.stack);
		console.log('');
		console.log(e.CL.trace());
		console.log('');
		throw e;
	};
	
	// ### CL.tag()
	// Set a custom tag for the Context that will be shown in the trace in case of unhandled error.
	CL.tag = function (tag) {
		current.tag = tag;
	};
	
	// ### CL.set(key, value)
	// Attach data to current context that can be accessed by any child contexts.
	// Can be used as kind of a global variable for session information etc.
	// As it is with global variables, must be used with care.
	CL.set = function (key, val) {
		current.set(key, val);
	};
	
	// ### CL.get(key)
	// Get a value by key previously defined by `CL.set()` on current context or one of its callers.
	CL.get = function (key) {
		if (current) return current.get(key);
	};
	
	// ### CL.env(data, Method|Context, args...)
	// Creates a new context with its data set to `data` and optionally runs Method with args.
	// If passed a Context instead of Method, will set its data to `data`.
	// Returns the created or modified Context.
	CL.env = function (data, arg) {
		// If it's a Context already, just replace data
		if (arg instanceof Context) {
			arg.data = data;
			return arg;
		}
		
		var cd = { data: data };
	
		if (isGenerator(arg)) {
			cd.generator = arg;
		}
		else {
			var l, args = null;
			if ((l = arguments.length - 2) > 0) {
				args = new ArrayClass(l);
				for (var i = 0; i < l; i++) {
					args[i] = arguments[i + 2];
				}
			}
			cd.method = arg;
			cd.args = args;
		}

		return push(new Context(cd));
	};
	
	var Value = CL.Value = function (value) {
		this.value = value;
	}
	// ### CL.val(value)
	// Creates a Value object that can be returned or yielded without evaluation.
	// Note that explicit `CL.wait()` or `CL.resolve()` will still resolve a context even if it's wraped in a Value.
	CL.val = function (value) {
		return new Value(value);
	}
	
	// ### yield CL.wait(what)
	// Waits on multiple `bg` tasks (passed as array or object) and returns said array/object with values filled in.
	CL.wait = CL(function* (obj) {
		if (obj instanceof Context || obj instanceof Value) {
			yield obj;
		}
		else if (isArray(obj)) {
			for (var i = 0, l = obj.length; i < l; i++) {
				if (obj[i] instanceof Context) {
					obj[i] = yield obj[i];
				}
			}
			return obj;
		}
		else if (obj && typeof obj == 'object') {
			if (obj[CL.key]) {
				yield obj[CL.key];
			}
			else {
				for (var key in obj) {
					if (hasOwnProperty(obj, key)) {
						if (obj[key] instanceof Context) {
							obj[key] = yield obj[key];
						}
					}
				}
			}
			return obj;
		}
		else {
			return obj;
		}
	});
	
	// ### yield CL.resolve(what)
	// Deep version of `CL.wait()`. Recursively waits on any Contexts found in an array or object.
	CL.resolve = CL(function* (obj) {
		// The recursive function must also be a generator to yield.
		// We do, however, use this one weird trick™ (`yield*`) to avoid creating unneeded Contexts.
		var rec = function* (obj) {
			if (isArray(obj)) {
				for (var i = 0, l = obj.length; i < l; i++) {
					if (obj[i] instanceof Value) {
						continue;
					}
					if (obj[i] instanceof Context) {
						obj[i] = yield obj[i];
					}
					if (obj[i] && typeof obj[i] == 'object') {
						yield* rec(obj[i]);
					}
				}
			}
			else {
				for (var key in obj) {
					if (hasOwnProperty(obj, key)) {
						if (obj[key] instanceof Value) {
							continue;
						}
						if (obj[key] instanceof Context) {
							obj[key] = yield obj[key];
						}
						if (obj[key] && typeof obj[key] == 'object') {
							yield* rec(obj[key]);
						}
					}
				}
			}
		};
		if (obj instanceof Value) return obj;
		if (obj instanceof Context) yield obj;
		else if (obj && typeof obj == 'object') yield* rec(obj);
		return obj;
	});
	
	// ### yield CL.sleep(ms);
	// Suspend the context for a said number of milliseconds.
	CL.sleep = CL(function* (ms) {
		var c = new Context({ working: true });
		setTimeoutFun(c.done.bind(c), ms);
		yield push(c);
	});
	
	// ### CL.expect()
	// Returns a Context that will return when `.done(value)` or `.throw(...)` is called on it.
	CL.expect = function () {
		return new Context({ working: true });
	};
	
	// ### yield CL.each(Array | Object, fun)
	// Run `fun(element, i)` over all elements of array
	// or `fun(value, key)` over all properties of an object.
	// For non-generator functions you should use something faster (e.g. Underscore)
	var eachArray = function* (arr, fun) {
		for (var i = 0, l = arr.length; i < l; i++) {
			yield fun(arr[i], i);
		}
	};
	var eachObject = function* (obj, fun) {
		for (var key in obj) {
			if (hasOwnProperty(obj, key)) yield fun(obj[key], key);
		}
	};
	CL.each = CL(function* (obj, fun) {
		yield isArray(obj) ? eachArray(obj, fun) : eachObject(obj, fun);
	});

	// ### yield CL.mapArray(Array | Object, fun)
	// Run `fun(element, i)` over all elements of array
	// or `fun(value, key)` over all properties of an object.
	// returns array or object with function's result values
	var mapArray = function* (arr, fun) {
		var result = [];
		for (var i = 0, l = arr.length; i < l; i++) {
			var res = yield fun(arr[i], i);
			result.push(res);
		}
		return yield result;
	};
	var mapObject = function* (obj, fun) {
		var result = {};
		for (var key in obj) {
			if (hasOwnProperty(obj, key)) {
				result[key] = yield fun(obj[key], key)
			}
		}
		return yield result;
	};
	CL.map = CL(function* (obj, fun) {
		return yield isArray(obj) ? mapArray(obj, fun) : mapObject(obj, fun);
	});
	
	// ### yield CL.parallel(Array | Object, fun, limit)
	//
	// * run `fun(element, i)` over all elements of array in parallel.
	// * run `fun(value, key)` over all properties of an object in parallel.
	//
	// Execution order is not guaranteed.
	// At most `limit` tasks will be running in parallel. The default limit is 1000.
	var parallelArray = function* (arr, fun, limit) {
		var r = 0;
		var self = current;
		var done = 0;
		for (var i = 0, l = arr.length; i < l; i++) {
			var c = push(new Context({ method: fun, args: [arr[i], i] }));
			c.handlers = [self];
			CL.bg(c);
			r++;
			while (r >= limit) {
				yield self.pause();
				r--;
				if (!(done++ % 10000)) self.clearBg();
			}
		}
	};
	var parallelObject = function* (obj, fun, limit) {
		var r = 0;
		var self = current;
		var done = 0;
		for (var key in obj) {
			if (hasOwnProperty(obj, key)) {
				var c = push(new Context({ method: fun, args: [obj[key], key] }));
				c.handlers = [self];
				CL.bg(c);
				r++;
				while (r >= limit) {
					yield self.pause();
					r--;
					if (!(done++ % 10000)) self.clearBg();
				}
			}
		}
	};
	CL.parallel = CL(function* (obj, fun, limit) {
		var lim = (limit === void 0) ? 1000 : limit;
		yield isArray(obj) ? parallelArray(obj, fun, lim) : parallelObject(obj, fun, lim);
	});

	// ### yield CL.mapParallel(Array | Object, fun, limit)
	//
	// * run `fun(element, i)` over all elements of array in parallel.
	// * run `fun(value, key)` over all properties of an object in parallel.
	// returns array or object with function's result values
	//
	// Execution order is not guaranteed.
	// At most `limit` tasks will be running in parallel. The default limit is 1000.
	var mapParallelArray = function* (arr, fun, limit) {
		var r = 0;
		var self = current;
		var done = 0;
		var result = [];
		for (var i = 0, l = arr.length; i < l; i++) {
			var c = push(new Context({ method: fun, args: [arr[i], i] }));
			result.push(c);
			c.handlers = [self];
			CL.bg(c);
			r++;
			while (r >= limit) {
				yield self.pause();
				r--;
				if (!(done++ % 10000)) self.clearBg();
			}
		}
		
		return yield CL.wait(result);
	};
	var mapParallelObject = function* (obj, fun, limit) {
		var r = 0;
		var self = current;
		var done = 0;
		var result = {};
		for (var key in obj) {
			if (hasOwnProperty(obj, key)) {
				var c = push(new Context({ method: fun, args: [obj[key], key] }));
				result[key] = c;
				c.handlers = [self];
				CL.bg(c);
				r++;
				while (r >= limit) {
					yield self.pause();
					r--;
					if (!(done++ % 10000)) self.clearBg();
				}
			}
		}
		return yield CL.wait(result);
	};
	CL.mapParallel = CL(function* (obj, fun, limit) {
		var lim = (limit === void 0) ? 1000 : limit;
		return yield isArray(obj) ? mapParallelArray(obj, fun, lim) : mapParallelObject(obj, fun, lim);
	});
	
	// ### CL.queue(options)
	// Constructs a Queue object with given options.
	CL.queue = function (options) {
		return new Queue(options || {});
	};
	
	// ## Queue
	// A class that allows queueing and limiting parallel execution of unrelated tasks.
	// Similar to host objects/contexts but a bit more flexible.
	var Queue = CL.Queue = function (options) {

		this.limit = options.limit || 1;
		this.sleep = options.sleep || 0;
		
		this.queue = [];
		this.running = 0;
		
	};
	
	Queue.prototype = {
	
		// ### Queue.push(task, args...)
		// Push a Task to queue. If task is a method, remaining arguments will be passed to it.
		push: function (task) {
			
			var c;
			if (building) {
				c = building;
				building = null;
			}
			else if (task instanceof Context) {
				c = task;
			}
			else if (isGenerator(task)) {
				c = push(new Context({
					generator: task
				}));
			}
			else {
				var cd = {
					method: task
				};
				var l = arguments.length - 1;
				if (l) {
					var args = new ArrayClass(l);
					for (var i = 0; i < l; i++) {
						args[i] = arguments[i + 1];
					}
					cd.args = args;
				}
				c = push(new Context(cd));
			}
			
			(c.handlers || (c.handlers = [])).push(this);
			
			if (this.running < this.limit) {
				this.running++;
				c.enqueue();
			}
			else {
				c.pause();
				this.queue.push(c);
			}
			
			return c;
			
		},
		
		// ### Queue.handleError()
		// Task error handler. Ignores error and moves to the next one
		handleError: function (c, e) {
			this.finish();
			return false;
		},
		
		// ### Queue.handleFinish()
		// Task success handler.
		handleFinish: function (c) {
			this.finish();
		},
		
		// ### Queue.finish()
		// Schedules moving on to the next task
		finish: function () {
			if (this.sleep) {
				var queue = this;
				setTimeout(function () {
					queue.next();
				}, this.sleep);
			}
			else {
				this.next();
			}
		},
		
		// ### Queue.next()
		// Runs next task or decrements the running counter
		next: function () {
			if (this.queue.length && this.running <= this.limit) {
				var c = this.queue.shift();
				c.resume();
			}
			else {
				this.running--;
			}
		}
		
	};
	
	// ### CL.debug(boolean)
	// Turn the debugging features on or off. Note that this may significantly degrade performance.
	CL.debug = function (set) {
		debug = set;
		if (!set) active = [];
	}
	
	// ### CL.getActive()
	// Get a list of Contexts that have started but haven't finished
	CL.getActive = function () {
		if (!debug) return false;
		cleanActive();
		return active;
	}
	
	// ### CL.log(boolean)
	// Turn the context log on or off. Note that this will prevent collection of memory until logging is disabled again.
	CL.log = function (set) {
		var had = log;
		log = set;
		if (had != set) logEntries = [];
	}
	
	CL.getLog = function () {
		return logEntries;
	}
	
	// # Footer
	// Assign ourselves to the global object (whichever it is) so only one CL is ever used.
	globalObject.CL = CL;
	if (typeof module != 'undefined') {
		// Export if CommonJS is used
		module.exports = CL;
	}
	else if (root) {
		// Not sure if there are cases when this is needed, but I leave it just in case.
		root.CL = CL;
	}
	
	// Check for a global flag that can be set before 
	if (typeof CLEnableDebugging != 'undefined') CL.debug(CLEnableDebugging);
	else if (typeof global == 'object') CL.debug(global.CLEnableDebugging);

})(this);
